/* -----------------------------------------------------------------------------
 *
 * DESCRIPTION:
 *	DOOM graphics stuff for Svgalib, Linux.
 *	Keyboard routine inspired from Snes9x - SVGALIB
 *
 *-----------------------------------------------------------------------------
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <stdarg.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <errnos.h>
#include <signal.h>

#include <vga.h>
#include <vgakeyboard.h>

#include "doomstat.h"
#include "i_system.h"
#include "v_video.h"

#include "m_argv.h"
#include "d_main.h"

#include "doomdef.h"
#include "r_draw.h"
#include "am_map.h"
#include "f_wipe.h"
#include "f_finale.h"
#include "r_main.h"

int		fbnum=0;

int		bpp;
int		pixel_size;
lighttable_t	truecolor_palette[256];
extern byte	*orig_colormaps;

static char prev_keystate [128] = "";

#ifdef DBLBUFFER
void *svgascreen[2];
#endif

// base of screen
unsigned char	*vgabase;

// Display functions
void		(*R_DrawColumn)();
void		(*R_DrawColumnLow)();
void		(*R_DrawFuzzColumn)();
void		(*R_DrawFuzzColumnLow)();
void		(*R_DrawTranslatedColumn)();
void		(*R_DrawTranslatedColumnLow)();
void		(*R_DrawSpan)();
void		(*R_DrawSpanLow)();
void		(*AM_DrawFline)();
void		(*V_DrawPatch)();
void		(*V_DrawPatchFlipped)();
int		(*wipe_doMelt)();
void		(*F_DrawPatchCol)();

extern	unsigned long fuzzmask;	// for fuzz effect in 15/16 bit mode


void I_ShutdownGraphics(void)
{
	keyboard_close();

	vga_setmode(TEXT);	
}

//
// I_StartFrame
//
void I_StartFrame (void)
{
}

//
// I_StartTic
//
void I_StartTic (void)
{
	event_t	event;

	// Taken from Snes9x

	char	*keystate;

	keyboard_update ();
	keystate=keyboard_getstate();

#define KEY_DOWN(a) (keystate[a])
#define KEY_PRESS(a) (keystate[a] && !prev_keystate[a])
#define KEY_WASPRESSED(a) (prev_keystate[a] && !keystate[a])
#define PROCESS_KEY(k,t)					\
	if (KEY_PRESS(k))					\
	{							\
			event.type = ev_keydown;		\
			event.data1 = t;			\
			D_PostEvent(&event);			\
	}							\
	if (KEY_WASPRESSED(k))					\
	{							\
			event.type = ev_keyup;			\
			event.data1 = t;			\
			D_PostEvent(&event);			\
	}

	PROCESS_KEY(SCANCODE_CURSORBLOCKRIGHT, KEY_RIGHTARROW);
	PROCESS_KEY(SCANCODE_CURSORBLOCKLEFT, KEY_LEFTARROW);
	PROCESS_KEY(SCANCODE_CURSORBLOCKDOWN, KEY_DOWNARROW);
	PROCESS_KEY(SCANCODE_CURSORBLOCKUP, KEY_UPARROW);

	PROCESS_KEY(SCANCODE_ESCAPE, KEY_ESCAPE);
	PROCESS_KEY(SCANCODE_ENTER, KEY_ENTER);
	PROCESS_KEY(SCANCODE_KEYPADENTER, KEY_ENTER);
	PROCESS_KEY(SCANCODE_TAB, KEY_TAB);
	PROCESS_KEY(SCANCODE_SPACE, ' ');
	PROCESS_KEY(SCANCODE_BREAK, KEY_PAUSE);
	PROCESS_KEY(SCANCODE_BREAK_ALTERNATIVE, KEY_PAUSE);

	PROCESS_KEY(SCANCODE_F1, KEY_F1);
	PROCESS_KEY(SCANCODE_F2, KEY_F2);
	PROCESS_KEY(SCANCODE_F3, KEY_F3);
	PROCESS_KEY(SCANCODE_F4, KEY_F4);
	PROCESS_KEY(SCANCODE_F5, KEY_F5);
	PROCESS_KEY(SCANCODE_F6, KEY_F6);
	PROCESS_KEY(SCANCODE_F7, KEY_F7);
	PROCESS_KEY(SCANCODE_F8, KEY_F8);
	PROCESS_KEY(SCANCODE_F9, KEY_F9);
	PROCESS_KEY(SCANCODE_F10, KEY_F10);
	PROCESS_KEY(SCANCODE_F11, KEY_F11);
	PROCESS_KEY(SCANCODE_F12, KEY_F12);
		
	PROCESS_KEY(SCANCODE_1, '1');
	PROCESS_KEY(SCANCODE_2, '2');
	PROCESS_KEY(SCANCODE_3, '3');
	PROCESS_KEY(SCANCODE_4, '4');
	PROCESS_KEY(SCANCODE_5, '5');
	PROCESS_KEY(SCANCODE_6, '6');
	PROCESS_KEY(SCANCODE_7, '7');
	PROCESS_KEY(SCANCODE_8, '8');
	PROCESS_KEY(SCANCODE_9, '9');
	PROCESS_KEY(SCANCODE_0, '0');

	PROCESS_KEY(SCANCODE_MINUS, KEY_MINUS);
	PROCESS_KEY(SCANCODE_KEYPADMINUS, KEY_MINUS);
	PROCESS_KEY(SCANCODE_EQUAL, KEY_EQUALS);
	PROCESS_KEY(SCANCODE_KEYPADPLUS, KEY_EQUALS);

	PROCESS_KEY(SCANCODE_BACKSPACE, KEY_BACKSPACE);
	PROCESS_KEY(SCANCODE_REMOVE, KEY_BACKSPACE);

	PROCESS_KEY(SCANCODE_LEFTSHIFT, KEY_RSHIFT);
	PROCESS_KEY(SCANCODE_RIGHTSHIFT, KEY_RSHIFT);
	PROCESS_KEY(SCANCODE_LEFTCONTROL, KEY_RCTRL);
	PROCESS_KEY(SCANCODE_RIGHTCONTROL, KEY_RCTRL);
	PROCESS_KEY(SCANCODE_LEFTALT, KEY_RALT);
	PROCESS_KEY(SCANCODE_RIGHTALT, KEY_RALT);

	PROCESS_KEY(SCANCODE_Q, 'q');
	PROCESS_KEY(SCANCODE_W, 'w');
	PROCESS_KEY(SCANCODE_E, 'e');
	PROCESS_KEY(SCANCODE_R, 'r');
	PROCESS_KEY(SCANCODE_T, 't');
	PROCESS_KEY(SCANCODE_Y, 'y');
	PROCESS_KEY(SCANCODE_U, 'u');
	PROCESS_KEY(SCANCODE_I, 'i');
	PROCESS_KEY(SCANCODE_O, 'o');
	PROCESS_KEY(SCANCODE_P, 'p');

	PROCESS_KEY(SCANCODE_A, 'a');
	PROCESS_KEY(SCANCODE_S, 's');
	PROCESS_KEY(SCANCODE_D, 'd');
	PROCESS_KEY(SCANCODE_F, 'f');
	PROCESS_KEY(SCANCODE_G, 'g');
	PROCESS_KEY(SCANCODE_H, 'h');
	PROCESS_KEY(SCANCODE_J, 'j');
	PROCESS_KEY(SCANCODE_K, 'k');
	PROCESS_KEY(SCANCODE_L, 'l');

	PROCESS_KEY(SCANCODE_Z, 'z');
	PROCESS_KEY(SCANCODE_X, 'x');
	PROCESS_KEY(SCANCODE_C, 'c');
	PROCESS_KEY(SCANCODE_V, 'v');
	PROCESS_KEY(SCANCODE_B, 'b');
	PROCESS_KEY(SCANCODE_N, 'n');
	PROCESS_KEY(SCANCODE_M, 'm');

	memcpy (prev_keystate, keystate, sizeof (prev_keystate));
}


//
// I_UpdateNoBlit
//
void I_UpdateNoBlit (void)
{
}

//
// I_FinishUpdate
//
extern patch_t	*ext_shortnum[10];

void I_FinishUpdate (void)
{

    static int	lasttic;
    int		tics;
    int		i;
	int c,d,u;

    // draws little dots on the bottom of the screen
    if (devparm)
    {

	i = I_GetTime();
	tics = i - lasttic;
	lasttic = i;

	if (tics>999) tics=999;
	c=tics / 100;
	d=(tics-c*100) / 10;
	u=(tics-c*100-d*10);

	if (ext_shortnum[c])
	{
		V_DrawPatch(320-12,0,0,ext_shortnum[c]);
		V_DrawPatch(320-8,0,0,ext_shortnum[d]);
		V_DrawPatch(320-4,0,0,ext_shortnum[u]);
	}
#if 0
	if (tics > 20) tics = 20;

	for (i=0 ; i<tics*2 ; i+=2)
	{
		switch (pixel_size)
		{
			case 1:
				screens[0][ (SCREENHEIGHT-1)*SCREENWIDTH + i] = 0xff;
				break;
			case 2:
				(unsigned short *) (screens[0])[ (SCREENHEIGHT-1)*SCREENWIDTH + i] = truecolor_palette[0xff];
				break;
			case 3:
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size] = truecolor_palette[0xff];
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size+1] = truecolor_palette[0xff] >> 8;
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size+2] = truecolor_palette[0xff] >> 16;
				break;
		}
	}
	for ( ; i<20*2 ; i+=2)
	{
		switch (pixel_size)
		{
			case 1:
				screens[0][ (SCREENHEIGHT-1)*SCREENWIDTH + i] = 0x00;
				break;
			case 2:
				(unsigned short *) (screens[0])[ (SCREENHEIGHT-1)*SCREENWIDTH + i] = truecolor_palette[0x00];
				break;
			case 3:
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size] = truecolor_palette[0x00];
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size+1] = truecolor_palette[0x00] >> 8;
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size+2] = truecolor_palette[0x00] >> 16;
				break;
		}
	}
#endif    
    }

	// Copie image finale + synchro VBL

#ifdef DBLBUFFER
	vga_copytoplanar256(svgascreen[fbnum],320,(fbnum*320*240)/4,80,320,240);
	vga_setdisplaystart(fbnum*320*240);
/*	vga_setpage(fbnum);*/
	vga_waitretrace();	

	fbnum =1-fbnum;
	screens[0]=svgascreen[fbnum]+320*20;
	R_InitBuffer(scaledviewwidth,viewheight); /* mettre a jour ylookup */
#else
	memcpy(vga_getgraphmem(),screens[0],SCREENWIDTH*SCREENHEIGHT*pixel_size);
#endif
}


//
// I_ReadScreen
//
void I_ReadScreen (byte* scr)
{
    memcpy (scr, screens[0], SCREENWIDTH*SCREENHEIGHT*pixel_size);
}


//
// Palette stuff.
//

//
// I_SetPalette
//
void I_SetPalette (byte* palette)
{
	int		i;
	lighttable_t	tc=0;
	byte		*palette1=palette;

	switch(pixel_size)
	{
		case 1:
			for(i = 0; i < 256; i++)
			{
				byte r,v,b;

				r = gammatable[usegamma][*palette++]>>2;
				v = gammatable[usegamma][*palette++]>>2;
				b = gammatable[usegamma][*palette++]>>2;

				vga_setpalette(i,r,v,b);
			}
			break;
		case 2:
			for(i = 0; i < 256; i++)
			{
				/* Palettes de lumiere */

				int		r1,v1,b1;
				int		c,y,r,v,b;

				c=orig_colormaps[i]*3;

				r1=gammatable[usegamma][palette[c]];
				v1=gammatable[usegamma][palette[c+1]];
				b1=gammatable[usegamma][palette[c+2]];	

				for(y=0;y<NUMCOLORMAPS;y++)
				{
					r=r1-(y*r1)/(NUMCOLORMAPS-1);
					v=v1-(y*v1)/(NUMCOLORMAPS-1);
					b=b1-(y*b1)/(NUMCOLORMAPS-1);	
					switch (bpp)
					{
						case 15:
							tc = ((r>>3)<<10)|((v>>3)<<5)|(b>>3);
							break;
						case 16:
							tc = ((r>>3)<<11)|((v>>2)<<5)|(b>>3);
							break;
					}
					colormaps[y*256+i]=tc<<16|tc;			
				}

				/* Invulnerabilite */
	
				colormaps[NUMCOLORMAPS*256+i]=colormaps[orig_colormaps[32*256+i]];

				/* Noir */
	
				colormaps[(NUMCOLORMAPS+1)*256+i]=0;

				/* Palettes normales */

				r=gammatable[usegamma][*palette1++];
				v=gammatable[usegamma][*palette1++];
				b=gammatable[usegamma][*palette1++];

				switch (bpp)
				{
					case 15:
						tc = ((r>>3)<<10)|((v>>3)<<5)|(b>>3);
						break;
					case 16:
						tc = ((r>>3)<<11)|((v>>2)<<5)|(b>>3);
						break;
				}
				truecolor_palette[i]=tc<<16|tc;
			}
			break;
		case 3:
			for(i = 0; i < 256; i++)
			{
				/* Palettes de lumiere */

				int		r1,v1,b1;
				int		c,y,r,v,b;

				c=orig_colormaps[i]*3;

				r1=gammatable[usegamma][palette[c]];
				v1=gammatable[usegamma][palette[c+1]];
				b1=gammatable[usegamma][palette[c+2]];	

				for(y=0;y<NUMCOLORMAPS;y++)
				{
					r=r1-(y*r1)/(NUMCOLORMAPS-1);
					v=v1-(y*v1)/(NUMCOLORMAPS-1);
					b=b1-(y*b1)/(NUMCOLORMAPS-1);	

					colormaps[y*256+i]=r<<16|v<<8|b;		
					
				}

				/* Invulnerabilite */
	
				colormaps[NUMCOLORMAPS*256+i]=colormaps[orig_colormaps[32*256+i]];

				/* Noir */
	
				colormaps[(NUMCOLORMAPS+1)*256+i]=0;

				/* Palettes normales */

				r=gammatable[usegamma][*palette1++];
				v=gammatable[usegamma][*palette1++];
				b=gammatable[usegamma][*palette1++];

				truecolor_palette[i]=r<<16|v<<8|b;
			}
			break;
		default:
			break;
	}
}

#ifdef DBLBUFFER
void ClearBuffer(char *monbuffer,int width,int height)
{
	int x,y;

	for (y=0;y<height;y++)
		for (x=0;x<width;x++)
			*monbuffer++=0;
}
#endif

void I_InitGraphics(void)
{
	static int	firsttime=1;
	int		mymode;

	if (!firsttime)
		return;
	firsttime = 0;

#ifdef DBLBUFFER
	svgascreen[0]=malloc(320*240);
	svgascreen[1]=malloc(320*240);
	screens[0]=svgascreen[1]+320*20;

	ClearBuffer(svgascreen[0],320,240);
	ClearBuffer(svgascreen[1],320,240);
#endif

	vga_init();

	signal(SIGINT, (void (*)(int)) I_Quit);

#ifdef DBLBUFFER
	mymode=G320x240x256;
#else
	mymode=G320x200x256;
#endif
	bpp=8;
#ifndef DBLBUFFER
	if (svga15) {
		mymode=G320x200x32K;
		bpp=15;
		if (!vga_hasmode(mymode))
			I_Error("Video : 15-bits screen depth not supported\n");
	}
	if (svga16) {
		mymode=G320x200x64K;
		bpp=16;
		if (!vga_hasmode(mymode))
			I_Error("Video : 16-bits screen depth not supported\n");
	}
	if (svga24) {
		mymode=G320x200x16M;
		bpp=24;
		if (!vga_hasmode(mymode))
			I_Error("Video : 24-bits screen depth not supported\n");
	}
#endif

	if (vga_setmode (mymode) < 0)
		I_Error ( "Unable to switch to requested screen mode/resolution:\n");

#ifdef DBLBUFFER
	vga_setdisplaystart(1);
#endif
	keyboard_init();

	switch (bpp)
	{
		case 8:
			R_DrawColumn=R_DrawColumn8;
			R_DrawColumnLow=R_DrawColumnLow8;
			R_DrawFuzzColumn=R_DrawFuzzColumn8;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow8;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn8;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow8;
			R_DrawSpan=R_DrawSpan8;
			R_DrawSpanLow=R_DrawSpanLow8;
			AM_DrawFline=AM_drawFline8;
			wipe_doMelt=wipe_doMelt8;
			V_DrawPatch=V_DrawPatch8;
			V_DrawPatchFlipped=V_DrawPatchFlipped8;
			F_DrawPatchCol=F_DrawPatchCol8;
			pixel_size=1;
			break;
		case 15:
			R_DrawColumn=R_DrawColumn16;
			R_DrawColumnLow=R_DrawColumnLow16;
			R_DrawFuzzColumn=R_DrawFuzzColumn16;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow16;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn16;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow16;
			R_DrawSpan=R_DrawSpan16;
			R_DrawSpanLow=R_DrawSpanLow16;
			AM_DrawFline=AM_drawFline16;
			wipe_doMelt=wipe_doMelt16;
			V_DrawPatch=V_DrawPatch16;
			V_DrawPatchFlipped=V_DrawPatchFlipped16;
			F_DrawPatchCol=F_DrawPatchCol16;
			pixel_size=2;
			fuzzmask=0x3DEF3DEF;
			break;
		case 16:
			R_DrawColumn=R_DrawColumn16;
			R_DrawColumnLow=R_DrawColumnLow16;
			R_DrawFuzzColumn=R_DrawFuzzColumn16;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow16;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn16;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow16;
			R_DrawSpan=R_DrawSpan16;
			R_DrawSpanLow=R_DrawSpanLow16;
			AM_DrawFline=AM_drawFline16;
			wipe_doMelt=wipe_doMelt16;
			V_DrawPatch=V_DrawPatch16;
			V_DrawPatchFlipped=V_DrawPatchFlipped16;
			F_DrawPatchCol=F_DrawPatchCol16;
			pixel_size=2;
			fuzzmask=0x7BEF7BEF;
			break;
		case 24:
			R_DrawColumn=R_DrawColumn24;
			R_DrawColumnLow=R_DrawColumnLow24;
			R_DrawFuzzColumn=R_DrawFuzzColumn24;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow24;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn24;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow24;
			R_DrawSpan=R_DrawSpan24;
			R_DrawSpanLow=R_DrawSpanLow24;
			AM_DrawFline=AM_drawFline24;
			wipe_doMelt=wipe_doMelt24;
			V_DrawPatch=V_DrawPatch24;
			V_DrawPatchFlipped=V_DrawPatchFlipped24;
			F_DrawPatchCol=F_DrawPatchCol24;
			pixel_size=3;
			break;
		default:
			I_Error("Unknown pixel size for screen\n");
	}
}

void I_SetMainScreen(void)
{
	/* screens[0] is copied to vga memory */
}
