/*  Emacs style mode select   -*- C++ -*-  */
/* ----------------------------------------------------------------------------- */
/*  */
/*  $Id:$ */
/*  */
/*  Copyright (C) 1993-1996 by id Software, Inc. */
/*  */
/*  This source is available for distribution and/or modification */
/*  only under the terms of the DOOM Source Code License as */
/*  published by id Software. All rights reserved. */
/*  */
/*  The source is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License */
/*  for more details. */
/*  */
/*  $Log:$ */
/*  */
/*  DESCRIPTION: */
/*  */
/* ----------------------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <stdarg.h>
#include <sys/time.h>
#include <unistd.h>

#ifdef ATARI
#include "atari.h"
#endif

#include "doomdef.h"
#include "m_misc.h"
#include "i_video.h"
#include "i_sound.h"

#include "d_net.h"
#include "g_game.h"

#ifdef __GNUG__
#pragma implementation "i_system.h"
#endif
#include "i_system.h"

int	mb_used = 8;	/*  8 Mo par dfaut */

void
I_Tactile
( int	on,
  int	off,
  int	total )
{
  /*  UNUSED. */
  on = off = total = 0;
}

ticcmd_t	emptycmd;
ticcmd_t*	I_BaseTiccmd(void)
{
    return &emptycmd;
}

#ifdef ATARI
int atari_HeapSize=8<<20;	/* 8 Mo par defaut */
#endif

int  I_GetHeapSize (void)
{
    return mb_used*1024*1024;
}

byte* I_ZoneBase (int*	size)
{
#ifdef ATARI
	*size=Mxalloc(-1,3)-131072;
	if (*size>atari_HeapSize)
		*size=atari_HeapSize;
#else
	*size = mb_used*1024*1024;
#endif

	printf(" %d Kbytes allocated for zone\n",(*size)>>10);

    return (byte *) malloc (*size);
}



/*  */
/*  I_GetTime */
/*  returns time in 1/70th second tics */
/*  */

#ifndef ATARI
int  I_GetTime (void)
{
    int			newtics;
    static int		basetime=0;
    struct timeval	tp;
    struct timezone	tzp;
  
    gettimeofday(&tp, &tzp);
    if (!basetime)
	basetime = tp.tv_sec;
    newtics = (tp.tv_sec-basetime)*TICRATE + tp.tv_usec*TICRATE/1000000;

    return newtics;
}
#endif


/*  */
/*  I_Init */
/*  */
void I_Init (void)
{
	I_InitSound();
	/* I_InitGraphics(); */
}

/*  */
/*  I_Quit */
/*  */
void I_Quit (void)
{
    D_QuitNetGame ();
    I_ShutdownSound();
    I_ShutdownMusic();
    M_SaveDefaults ();
    I_ShutdownGraphics();
    exit(0);
}

void I_WaitVBL(int count)
{
#ifdef SGI
    sginap(1);                                           
#else
#ifdef SUN
    sleep(0);
#else
#ifdef ATARI
	int i;

	for (i=0;i<count;i++)
		Vsync();
#else
    usleep (count * (1000000/70) );                                
#endif
#endif
#endif
}

void I_BeginRead(void)
{
}

void I_EndRead(void)
{
}

byte*	I_AllocLow(int length)
{
    byte*	mem;
        
    mem = (byte *)malloc (length);
    memset (mem,0,length);
    return mem;
}


/*  */
/*  I_Error */
/*  */
extern boolean demorecording;

void I_Error (char *error, ...)
{
    va_list	argptr;

    /*  Message first. */
    va_start (argptr,error);
    fprintf (stderr, "Error: ");
    vfprintf (stderr,error,argptr);
    fprintf (stderr, "\n");
    va_end (argptr);

    fflush( stderr );

    /*  Shutdown. Here might be other errors. */
    if (demorecording)
	G_CheckDemoStatus();

    D_QuitNetGame ();
    I_ShutdownGraphics();
    
    exit(-1);
}
