/* DESCRIPTION: */
/* 	DOOM graphics stuff for Atari Falcon/TT Xbios. */
/*  */

#include <signal.h>

#include "atari.h"

#include "doomstat.h"
#include "i_system.h"
#include "v_video.h"

#include "m_argv.h"
#include "d_main.h"

#include "doomdef.h"
#include "r_local.h"
#include "am_map.h"
#include "f_wipe.h"
#include "f_finale.h"

#include "z_zone.h"

static int	VideoDone=0;	/* Video initialized ? */

int fbnum;	/* Screen buffer to draw to */

int		pixel_size;
int		bpp;
lighttable_t	truecolor_palette[256];
extern byte	*orig_colormaps;

extern	unsigned long fuzzmask;	/*  for fuzz effect in 15/16 bit mode */

unsigned short	atari_oldvmode;
void		*atari_oldvbase;
unsigned long	atari_screen[2];	/* game screen */
void		*atari_ecralloc[2];
unsigned long	atari_couleur0;

unsigned long	atari_offset;	/* offset for game screen in real screen */

extern void I_InitInterrupts();
extern void I_ShutdownInterrupts();
event_t atari_event;
extern short	atari_CtrlCPressed;

int	cookie_mch=0;
int	cookie_vdo=0;

/*  Display functions */
void		(*R_DrawColumn)();
void		(*R_DrawColumnLow)();
void		(*R_DrawFuzzColumn)();
void		(*R_DrawFuzzColumnLow)();
void		(*R_DrawTranslatedColumn)();
void		(*R_DrawTranslatedColumnLow)();
void		(*R_DrawSpan)();
void		(*R_DrawSpanLow)();
void		(*AM_DrawFline)();
int		(*wipe_doMelt)();
void		(*V_DrawPatch)();
void		(*V_DrawPatchFlipped)();
void		(*F_DrawPatchCol)();

/* Chunky to planar conversion */

void init_c2p(void);
void convert_c2p(byte *src,byte *dest,long width,long height,long dblligne);

void ScanCookies(void)
{
	void *oldpile;
	unsigned long *cookie_jar;

	char *atari_machine[4]={
		"Atari ST",
		"Atari STE/ST Book/Mega STE",
		"Atari TT",
		"Atari Falcon 030"
	};

	char *atari_video[4]={
		"ST Shifter",
		"STE Shifter",
		"TT Shifter",
		"Falcon Videl"
	};

	oldpile=(void *)Super(0);
	cookie_jar=_P_COOKIES;
	Super(oldpile);	

	printf(" Reading hardware configuration...\n");

	while (*cookie_jar)
	{
		if (*cookie_jar==COOKIE_MCH)
		{
			cookie_mch=(*(cookie_jar+1))>>16;
			printf("  Machine : %s\n",atari_machine[cookie_mch]);
			if ((cookie_mch!=MCH_TT) && (cookie_mch!=MCH_F30))
			{
				I_Error("You need a TT or a Falcon\n");
			}			
		}

		if (*cookie_jar==COOKIE_VDO)
		{
			cookie_vdo=(*(cookie_jar+1))>>16;
			printf("  Video : %s\n",atari_video[cookie_vdo]);
			if ((cookie_vdo!=VDO_TT) && (cookie_vdo!=VDO_F30))
			{
				I_Error("You need a TT or a Falcon\n");
			}			
		}

		cookie_jar+=2;
	}
}

void I_AllocateScreensAtStart(void)
{
	int width,height;
	int i,x,y;
	short *myscreen;

	ScanCookies();

	width=SCREENWIDTH;
	height=SCREENHEIGHT;

	/* --- Read video properties --- */

	switch (cookie_vdo)
	{
		case VDO_TT:
			height=240;  	/* 480 8bits <-> 240 16 bits */
			atari_offset=SCREENWIDTH*40*2;
			break;
		case VDO_F30:
			switch (Montype())
			{
				case 0:		/* MONO */
					I_Error("Doom does not work on a monochrome monitor\n");
					break;
				case 1:		/* RVB : 320x200 */
				case 3:		/* TV : 320x200 */
					atari_offset=0;
					break;
				case 2:		/* VGA : 320x240 */
					atari_offset=SCREENWIDTH*20*2;
					height=240;
					break;
			}
			break;
	}

	/* --- Allocate screens */

	for (i=0;i<2;i++)
	{
		atari_ecralloc[i]=Mxalloc((width*height*2)+256,0);

		if (atari_ecralloc[i]==NULL)
			I_Error("Not enough memory for screens.\n");

		/* Clear the allocated screen */

		myscreen=atari_ecralloc[i];
		for (y=0;y<height;y++)
			for (x=0;x<SCREENWIDTH;x++)
				*myscreen++=0;

		atari_screen[i]=(unsigned long )(atari_ecralloc[i]+16UL) & 0xFFFFFFF0UL;
        }
}

void I_ShutdownGraphics(void)
{
	void	*oldpile;

	if (VideoDone==-1)
	{
		Supexec(I_ShutdownInterrupts);

		switch(cookie_vdo)
		{
			case VDO_TT:
				Setscreen(-1,atari_oldvbase,-1,-1);
				EsetShift(atari_oldvmode);
				break;
			case VDO_F30:
				/* Bordure noire */

				oldpile=(void *)Super(0);
				VIDEL_COLOR0=atari_couleur0;
				Super(oldpile);

				Setscreen(-1,atari_oldvbase,-1,-1);
			        Vsetmode(atari_oldvmode);
				break;
		}

		Mfree(atari_ecralloc[0]);
		Mfree(atari_ecralloc[1]);

		VideoDone=0;
	}
}



/*  */
/*  I_StartFrame */
/*  */
void I_StartFrame (void)
{
}

/*  */
/*  I_StartTic */
/*  */
void I_StartTic (void)
{
	if (atari_CtrlCPressed!=0)
		I_Quit();
}


/*  */
/*  I_UpdateNoBlit */
/*  */
void I_UpdateNoBlit (void)
{
}

/*  */
/*  I_FinishUpdate */
/*  */
extern patch_t	*ext_shortnum[10];

void I_FinishUpdate (void)
{
/*	unsigned short	*destscr; */

    static int	lasttic;
    int		tics;
    int		i;
	int c,d,u;

    /*  draws little dots on the bottom of the screen */
    if (devparm)
    {

	i = I_GetTime();
	tics = i - lasttic;
	lasttic = i;

	if (tics>999) tics=999;
	c=tics / 100;
	d=(tics-c*100) / 10;
	u=(tics-c*100-d*10);

	if (ext_shortnum[c])
	{
		V_DrawPatch(320-12,0,0,ext_shortnum[c]);
		V_DrawPatch(320-8,0,0,ext_shortnum[d]);
		V_DrawPatch(320-4,0,0,ext_shortnum[u]);
	}

#if 0
	if (tics > 20) tics = 20;

	for (i=0 ; i<tics*2 ; i+=2)
	{
		switch (pixel_size)
		{
			case 1:
				screens[0][ (SCREENHEIGHT-1)*SCREENWIDTH + i] = 0xff;
				break;
			case 2:
				destscr=(unsigned short *) screens[0];
				destscr[(SCREENHEIGHT-1)*SCREENWIDTH + i] = truecolor_palette[0xff];
				break;
			case 3:
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size] = truecolor_palette[0xff];
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size+1] = truecolor_palette[0xff] >> 8;
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size+2] = truecolor_palette[0xff] >> 16;
				break;
		}
	}
	for ( ; i<20*2 ; i+=2)
	{
		switch (pixel_size)
		{
			case 1:
				screens[0][ (SCREENHEIGHT-1)*SCREENWIDTH + i] = 0x00;
				break;
			case 2:
				destscr=(unsigned short *) screens[0];
				destscr[(SCREENHEIGHT-1)*SCREENWIDTH + i] = truecolor_palette[0x00];
				break;
			case 3:
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size] = truecolor_palette[0x00];
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size+1] = truecolor_palette[0x00] >> 8;
				screens[0][ ((SCREENHEIGHT-1)*SCREENWIDTH + i)*pixel_size+2] = truecolor_palette[0x00] >> 16;
				break;
		}
	}
#endif    
    }

	/* Permutation des ecrans */

	if (pixel_size==1)
	{
		convert_c2p(screens[0],(byte *)(atari_screen[fbnum]+atari_offset),320,200,(cookie_vdo==VDO_TT));
	}

	Setscreen(-1,atari_screen[fbnum],-1,-1);
	fbnum = 1-fbnum;				
	if (pixel_size!=1)
	{
		screens[0]=(byte *)(atari_screen[fbnum]+atari_offset);
		R_InitBuffer(scaledviewwidth,viewheight); /* mettre a jour ylookup */
	}
}


/*  */
/*  I_ReadScreen */
/*  */
void I_ReadScreen (byte* scr)
{
    memcpy (scr, screens[0], SCREENWIDTH*SCREENHEIGHT*pixel_size);
}

/*  */
/*  I_SetPalette */
/*  */
void I_SetPalette(byte *palette)
{
	int		i;
	byte		*palette1=palette;
	unsigned short	tc=0;
	unsigned short	TT_palette[256];
	unsigned long	F30_palette[256];

	switch(pixel_size)
	{
		case 1:
			palette1=palette;
			switch(cookie_vdo)
			{
				case VDO_TT:
					for (i=0;i<256;i++)
					{
						int r,v,b;

						r = gammatable[usegamma][*palette++];	
						v = gammatable[usegamma][*palette++];	
						b = gammatable[usegamma][*palette++];	
					
						tc=((r>>4)<<8)|((v>>4)<<4)|(b>>4);
						TT_palette[i]=tc;
					}
					EsetPalette(0,256,&TT_palette[0]);
					break;
				case VDO_F30:
					for(i = 0; i < 256; i++)
					{
						int r,v,b;

						r = gammatable[usegamma][*palette++];	
						v = gammatable[usegamma][*palette++];
						b = gammatable[usegamma][*palette++];

						F30_palette[i]=(r<<16)|(v<<8)|b;
					}
					VsetRGB(0,256,&F30_palette[0]);
					break;
			}
			/* Remplir les colormaps avec les valeurs des octets */
			for (i=0;i<256;i++)
			{
				int y;

				for (y=0;y<NUMCOLORMAPS+2;y++)
				{
					unsigned long c;

					c=colormaps[y*256+i] & 255;

					c|=(c<<24)|(c<<16)|(c<<8);

					colormaps[y*256+i]=c;
				}
			}
			break;
		case 2:
			for(i = 0; i < 256; i++)
			{
				/* Palettes de lumiere */

				int		r1,v1,b1;
				int		c,y,r,v,b;

				c=orig_colormaps[i]*3;

				r1=gammatable[usegamma][palette[c]];
				v1=gammatable[usegamma][palette[c+1]];
				b1=gammatable[usegamma][palette[c+2]];	

				for(y=0;y<NUMCOLORMAPS;y++)
				{
					r=r1-(y*r1)/(NUMCOLORMAPS-1);
					v=v1-(y*v1)/(NUMCOLORMAPS-1);
					b=b1-(y*b1)/(NUMCOLORMAPS-1);	
					switch (bpp)
					{
						case 15:
							tc = ((r>>3)<<10)|((v>>3)<<5)|(b>>3);
							break;
						case 16:
							tc = ((r>>3)<<11)|((v>>2)<<5)|(b>>3);
							break;
					}
					colormaps[y*256+i]=tc<<16|tc;			
				}

				/* Invulnerabilite */
	
				colormaps[NUMCOLORMAPS*256+i]=colormaps[orig_colormaps[32*256+i]];

				/* Noir */
	
				colormaps[(NUMCOLORMAPS+1)*256+i]=0;

				/* Palettes normales */

				r=gammatable[usegamma][*palette1++];
				v=gammatable[usegamma][*palette1++];
				b=gammatable[usegamma][*palette1++];

				switch (bpp)
				{
					case 15:
						tc = ((r>>3)<<10)|((v>>3)<<5)|(b>>3);
						break;
					case 16:
						tc = ((r>>3)<<11)|((v>>2)<<5)|(b>>3);
						break;
				}
				truecolor_palette[i]=tc<<16|tc;
			}
			break;
		case 3:
			for(i = 0; i < 256; i++)
			{
				/* Palettes de lumiere */

				int		r1,v1,b1;
				int		c,y,r,v,b;

				c=orig_colormaps[i]*3;

				r1=gammatable[usegamma][palette[c]];
				v1=gammatable[usegamma][palette[c+1]];
				b1=gammatable[usegamma][palette[c+2]];	

				for(y=0;y<NUMCOLORMAPS;y++)
				{
					r=r1-(y*r1)/(NUMCOLORMAPS-1);
					v=v1-(y*v1)/(NUMCOLORMAPS-1);
					b=b1-(y*b1)/(NUMCOLORMAPS-1);	

					colormaps[y*256+i]=r<<16|v<<8|b;		
					
				}

				/* Invulnerabilite */
	
				colormaps[NUMCOLORMAPS*256+i]=colormaps[orig_colormaps[32*256+i]];

				/* Noir */
	
				colormaps[(NUMCOLORMAPS+1)*256+i]=0;

				/* Palettes normales */

				r=gammatable[usegamma][*palette1++];
				v=gammatable[usegamma][*palette1++];
				b=gammatable[usegamma][*palette1++];

				truecolor_palette[i]=r<<16|v<<8|b;
			}
			break;
	}
}

void I_InitGraphics(void)
{
        short new_mode;
	void *oldpile;

	fbnum=0;

	init_c2p();

	switch(cookie_vdo)
	{
		case VDO_TT:
			atari_oldvbase=Logbase();
			atari_oldvmode=EgetShift();

			Setscreen(-1,atari_screen[1],-1,-1);
			EsetShift(TT_LOW);

			bpp=8;

			break;
		case VDO_F30:
			atari_oldvmode=Vsetmode(-1);
			atari_oldvbase=Logbase();	

			new_mode=atari_oldvmode & ( VGA + PAL );
			if (falcon8)
			{
				new_mode|=BPS8;
				atari_offset>>=1; /* diviser par 2 */
				bpp=8;
			} else {
				new_mode|=BPS16;
				bpp=16;
			}
			if (Montype()==2)
				new_mode|=VERTFLAG;

			Setscreen(-1,atari_screen[1],-1,-1);
			Vsetmode(new_mode);

			/* Bordure noire */

			oldpile=(void *)Super(0);
			atari_couleur0=VIDEL_COLOR0;
			VIDEL_COLOR0=NULL;
			Super(oldpile);

			break;
	}


	/* Interruptions */

	Supexec(I_InitInterrupts);

	/* OK */

	VideoDone=-1;


    signal(SIGINT, (void (*)(int)) I_Quit);

	switch (bpp)
	{
		case 8:
			R_DrawColumn=R_DrawColumn8;
			R_DrawColumnLow=R_DrawColumnLow8;
			R_DrawFuzzColumn=R_DrawFuzzColumn8;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow8;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn8;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow8;
			R_DrawSpan=R_DrawSpan8;
			R_DrawSpanLow=R_DrawSpanLow8;
			AM_DrawFline=AM_drawFline8;
			wipe_doMelt=wipe_doMelt8;
			V_DrawPatch=V_DrawPatch8;
			V_DrawPatchFlipped=V_DrawPatchFlipped8;
			F_DrawPatchCol=F_DrawPatchCol8;
			pixel_size=1;
			break;
		case 15:
			R_DrawColumn=R_DrawColumn16;
			R_DrawColumnLow=R_DrawColumnLow16;
			R_DrawFuzzColumn=R_DrawFuzzColumn16;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow16;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn16;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow16;
			R_DrawSpan=R_DrawSpan16;
			R_DrawSpanLow=R_DrawSpanLow16;
			AM_DrawFline=AM_drawFline16;
			wipe_doMelt=wipe_doMelt16;
			V_DrawPatch=V_DrawPatch16;
			V_DrawPatchFlipped=V_DrawPatchFlipped16;
			F_DrawPatchCol=F_DrawPatchCol16;
			pixel_size=2;
			fuzzmask=0x3DEF3DEF;
			break;
		case 16:
			R_DrawColumn=R_DrawColumn16;
			R_DrawColumnLow=R_DrawColumnLow16;
			R_DrawFuzzColumn=R_DrawFuzzColumn16;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow16;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn16;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow16;
			R_DrawSpan=R_DrawSpan16;
			R_DrawSpanLow=R_DrawSpanLow16;
			AM_DrawFline=AM_drawFline16;
			wipe_doMelt=wipe_doMelt16;
			V_DrawPatch=V_DrawPatch16;
			V_DrawPatchFlipped=V_DrawPatchFlipped16;
			F_DrawPatchCol=F_DrawPatchCol16;
			pixel_size=2;
			fuzzmask=0x7BEF7BEF;
			break;
		case 24:
			R_DrawColumn=R_DrawColumn24;
			R_DrawColumnLow=R_DrawColumnLow24;
			R_DrawFuzzColumn=R_DrawFuzzColumn24;
			R_DrawFuzzColumnLow=R_DrawFuzzColumnLow24;
			R_DrawTranslatedColumn=R_DrawTranslatedColumn24;
			R_DrawTranslatedColumnLow=R_DrawTranslatedColumnLow24;
			R_DrawSpan=R_DrawSpan24;
			R_DrawSpanLow=R_DrawSpanLow24;
			AM_DrawFline=AM_drawFline24;
			wipe_doMelt=wipe_doMelt24;
			V_DrawPatch=V_DrawPatch24;
			V_DrawPatchFlipped=V_DrawPatchFlipped24;
			F_DrawPatchCol=F_DrawPatchCol24;
			pixel_size=3;
			break;
		default:
			I_Error("Unknown pixel size for screen\n");
	}
}

void I_SetMainScreen(void)
{
	/* Put screens[0] on our screen */

	if (pixel_size!=1)
		screens[0]=(byte *)(atari_screen[1]+atari_offset);
}
